# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko


class ShellError(tobiko.TobikoException):
    pass


class ShellCommandFailed(ShellError):
    """Raised when shell command exited with non-zero status
    """
    message = ("command {command!r} failed (exit status is {exit_status}); "
               "stderr:\n{stderr!s}\n"
               "stdout:\n{stdout!s}")


class ShellTimeoutExpired(ShellError):
    """Raised when shell command timeouts and has been killed before exiting
    """
    message = ("command {command!r} timed out after {timeout!s} seconds; "
               "stderr:\n{stderr!s}\n"
               "stdout:\n{stdout!s}")
